# EnSimpleStaging  |  Copyright(C), 2004, Enfold Systems, LLC
# see LICENSE.txt for details

from __future__ import nested_scopes

import sys
import posixpath
import transaction
from difflib import SequenceMatcher
from sets import Set
from zope.interface import providedBy, directlyProvides, directlyProvidedBy

from DateTime import DateTime
from BTrees.IIBTree import intersection, difference as i_difference
from BTrees.IOBTree import IOBTree
from BTrees.OIBTree import OITreeSet, union, difference as o_difference, intersection as o_intersection
from BTrees.Length import Length

from AccessControl import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_base, aq_acquire
from Products.Archetypes.public import *
from Products.Archetypes.utils import findDict
from Products.ATContentTypes.content.folder import ATFolder
from Products.CMFCore.utils import getToolByName
from Products.CMFStaging.StagingTool import StageObjects
from Products.EnSimpleStaging.config import PROJECTNAME
from Products.EnSimpleStaging.event import publishStart
from Products.EnSimpleStaging.interfaces import IInPlaceStaging
from Products.AdvancedQuery import Eq, Generic, Or, In, Ge, Le
from Products.AdvancedQuery.eval import eval as evalAdvancedQuery

try:
    from Products.CMFCore.permissions import View, ReviewPortalContent
except ImportError:
    from Products.CMFCore.CMFCorePermissions import View, ReviewPortalContent

# JobServer hook:
try:
    from Products.JobServer.interfaces import IJobQueue
except ImportError:
    IJobQueue = None

from Products.PythonScripts.standard import urlencode
from ZEO.Exceptions import ClientDisconnected

try:
    from entransit.common.exceptions import DeploymentError, \
                                            DeploymentInProgress, OSError, \
                                            DependencyError, URLRewritingError
except ImportError:
    # Only DeploymentError is needed, because if we declare it here it means
    # it will not be raised anywere, and the others are only used inside the
    # try except that catches it.
    class DeploymentError(Exception):
        pass

# This is for the new HTTP transport code
try:
    from entransit.common.exceptions import ConnectionError, \
                                            RemoteDeploymentError
except ImportError:
    class ConnectionError(Exception):
        pass
    class RemoteDeploymentError(Exception):
        pass

class StagingArea(OrderedBaseFolder):
    """ A simple workspace implementation """

    security = ClassSecurityInfo()
    portal_type = 'StagingArea'
    meta_type = archetype_name = 'Staging Area'
    _at_rename_after_creation = True # rename object according to the title?
    _at_type_subfolder = 'Folder'

    allow_discussion = False
    filter_content_types  = 0
    allowed_content_types = ("Sandbox", )
    content_icon = "staging-area-icon.gif"

    actions = (#{'id': 'view',
               # 'name': 'View',
               # 'action': 'string:${object_url}/stagingarea_view',
               # 'permissions': (View,)},
               #{'id': 'stage_map',
               # 'name': 'Stage Map',
               # 'action': 'string:${object_url}/changeset_tree',
               # 'permissions': (ReviewPortalContent,)},
               {'id': 'deployment',
                'name': 'Deployment',
                'action': 'string:${object_url}/workspace_deployment',
                'permissions': (ReviewPortalContent,)},
               )

    _stage_info = None
    _si_length = None

    schema = BaseFolder.schema + Schema((

        StringField(
            name='stagePath',
            write_permission=ReviewPortalContent,
            widget=StringWidget(
                label='Deployment Path',
                condition='not:object/doInPlace',
                visible={'edit': 'visible', 'view': 'invisible'},
                description=(
                    'Path relative to the portal object, '
                    'where content will be deployed into. '
                    'I.e. public_website/sales '
                    'NOTICE: the destination must exist. '
                    'An empty value means this workspace is disabled and '
                    'will not deploy at all.'
                ),
            ),
        ),

        BooleanField(
            name='queueJobsByDefault',
            write_permission=ReviewPortalContent,
            default=False,
            widget=BooleanWidget(
                label='Queue deployments by default?',
                visible={'edit': 'visible', 'view': 'invisible'},
                description=(
                    'This switch only affects the default value of the '
                    '"Queued?" checkbox on the Deployment tab. '
                    '"Queueing" a deployment means that the deployment '
                    'is not done immediately, but is queued into the '
                    'JobServer job queue. Of course, this only happens '
                    'if the job queue is installed.'
                ),
            ),
        ),

        BooleanField(
            name='stagingUseCatalogOnly',
            write_permission=ReviewPortalContent,
            default=True,
            widget=BooleanWidget(
                label='Use Catalog for computing Staging Area change sets?',
                visible={'edit': 'visible', 'view': 'invisible'},
                description=(
                    'This switch can be turned on if the staging area '
                    'is small and contains non-Plone content '
                    'to enable picking up add/remove for those items.'
                ),
            ),
        ),

        LinesField(
            name='deployStates',
            write_permission=ReviewPortalContent,
            accessor='getDeployStates',
            mutator='setDeployStates',
            default=(),
            widget=LinesWidget(
                label='Deploy States',
                visible={'edit': 'visible', 'view': 'invisible'},
                description=(
                    'Only deploy objects in the following workflow states:'
                ),
            ),
        ),

        LinesField(
            name='typesWhiteList',
            write_permission=ReviewPortalContent,
            default=(),
            vocabulary='_ess_vocabularyAllTypes',
            widget=InAndOutWidget(
                label='Portal Types whitelist',
                visible={'edit': 'visible', 'view': 'invisible'},
                description=(
                    'If any portal_type is selected here, only the selected '
                    'types can be added inside the workspace. If nothing is '
                    'listed here, no filtering is done'
                ),
            ),
        ),

    ))

    def _resetAnnotations(self):
        self._stage_info = IOBTree()
        self._si_length = Length()

    def initializeArchetype(self, **kwargs):
        BaseFolder.initializeArchetype(self, **kwargs)
        self._resetAnnotations()

    security.declarePublic('doInPlace')
    def doInPlace(self):
        """ returns whether this workspace is in in-place mode or not """
        # Backwards compatibility
        bw = getattr(aq_base(self), 'inPlace', None)
        if bw is not None:
            return bw
        return IInPlaceStaging.providedBy(self)

    def setInPlace(self, value):
        provided = directlyProvidedBy(self)
        if not value:
            if IInPlaceStaging.providedBy(self):
                # Do we provide it?
                if IInPlaceStaging in provided:
                    # Marker interface, asking for removal, remove it
                    directlyProvides(self, provided - IInPlaceStaging)
                else:
                    # Might be set on the class, not much we can do.
                    return
            else:
                # Doensn't provide it, noop.
                return
        else:
            # Setting to inPlace.
            if not IInPlaceStaging.providedBy(self):
                # Does not provide yet, set using directlyProvides
                directlyProvides(self, provided, IInPlaceStaging)
            else:
                # Already provides it, skip.
                return


    security.declareProtected(ReviewPortalContent, 'setStagePath')
    def setStagePath(self, value):
        """ Sets the destination stage
        """
        # change the destination stage information
        field = self.Schema()['stagePath']
        value = str(value) # in case of unicode/marshall
        StringField.set(field, self, value)
        if not value:
            # the stage has been disabled for deployment
            return
        wt = getToolByName(self, 'portal_workspaces')
        wt.checkDuplicateStages()

    def _getSourceStageName(self):
        """ Source Stage Identifier 
        """
        return 'src_%s' % self.UID()
    getSourceStageName = _getSourceStageName

    security.declarePrivate('ess_filterWhitelistedTypes')
    def ess_filterWhitelistedTypes(self, portal_types):
        """ filters what object types can actually be added inside
        the portal """
        whitelist = self.getTypesWhiteList()
        # if whitelist is empty, do not filter
        if not whitelist:
            return portal_types

        from sets import Set
        whitelist = Set(whitelist)
        filtered_portal_types = [ptype for ptype in portal_types
                                 if ptype.getId() in whitelist]
        return filtered_portal_types


    security.declarePrivate('_ess_vocabularyAllTypes')
    def _ess_vocabularyAllTypes(self):
        portal_types = getToolByName(self, 'portal_types')
        typelist = [(fti.title_and_id(), fti.getId())
                     for fti in portal_types.listTypeInfo()]
        typelist.sort()
        return DisplayList([(id, title) for title, id in typelist])

    security.declarePrivate('getSourceStageInfo')
    def getSourceStageInfo(self):
        ut = getToolByName(self, 'portal_url')
        title = self.title_or_id() + " (source)"
        return (self._getSourceStageName(), # id
                title, # title (duh :-)
                ut.getRelativeContentURL(self), # path
                )

    security.declarePrivate('getDestinationStageInfo')
    def getDestinationStageInfo(self):
        field = self.Schema()['stagePath']
        path = StringField.get(field, self)
        if not path:
            # no dest. path initialized in this object
            return None
        portal = getToolByName(self, 'portal_url').getPortalObject()
        title = self.title_or_id() + " (destination)"
        return ('dst_%s' % self.UID(), # id
                title,
                path,
                )

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        BaseFolder.manage_afterAdd(self, item, container)
        wt = getToolByName(self, 'portal_workspaces')
        wt.checkDuplicateStages()

    def _globalLabel(self, label):
        # global labels must be unique among all labels in a site
        prefix = self.UID() + '-'
        if label.startswith(prefix):
            # It's already uniquefied
            return label
        return prefix + label

    security.declareProtected(StageObjects, 'getNextLabel')
    def getNextLabel(self):
        return "%s" % self._si_length()

    def getLastChangedPaths(self, query=None, 
                            applyFilter=False, 
                            filterChanged=True):
        return self.getDeploymentCandidates(query=query, 
                                            applyFilter=applyFilter,
                                            filterChanged=filterChanged)

    def getDeploymentCandidates(self, query=None, 
                                applyFilter=False, 
                                filterChanged=False,
                                adds=None):
        stagePath = '/'.join(self.getPhysicalPath())
        pathFilter = Eq('path', stagePath)

        if not query:
            query = pathFilter
        else:
            query &= pathFilter

        # Since the user doing the publication here will always have
        # 'AccessInactivePortalContent' permission, we have to do the
        # manual filtering ourselves, because we are selecting content
        # that will be deployed to the public website. Otherwise, we
        # might end up selecting expired content for deployment.
        ct = getToolByName(self, 'portal_catalog')
        now = ct.ZopeTime()
        if 'ValidityRange' in ct.Indexes.objectIds():
            query &= Eq('ValidityRange', now)
        elif 'effectiveRange' in ct.Indexes.objectIds():
            query &= Eq('effectiveRange', now)
        else:
            query &= Le('effective', now)
            query &= Ge('expires', now)

        if applyFilter:
            deploy_states = self.getDeployStates()
            if deploy_states:
                query &= In('review_state', deploy_states)

        if filterChanged:
            # return paths for objects from the catalog
            # that were changed since the last publication or tagging
            annotIndex, annot = self.getLastReversibleAnnotation() 
            lastPubDate = annot['date']

            dateFilter = Generic('lastModified', {'query': lastPubDate,
                                                  'range': 'min'})
            changedFilter = Eq('isResourceChanged', True)
            
            if applyFilter:
                query = query & (dateFilter | changedFilter)
            else:
                query = query & dateFilter
            
        changed = evalAdvancedQuery(ct, query)

        changed = OITreeSet([brain.getPath() for brain in changed
                             # avoid the stage itself
                             if brain.getPath() != stagePath
                             # avoid CompositePack attributes and its subobjects.
                             and 'cp_container' not in brain.getPath()
                             # avoid portal_factory leftovers
                             and 'portal_factory' not in brain.getPath()
                             ])

        # changed = set(self.ess_getLastChangedPathsSince(lastPubDate, 
        #                                                 query=query))

        if not self.getStagingUseCatalogOnly():
            wt = getToolByName(self, 'portal_workspaces')
            repo = wt._getVersionRepository()
            def collect(item):
                path = '/'.join(item.getPhysicalPath())
                if not repo.isUnderVersionControl(item) and adds is not None:
                    adds.insert(path)
                if (not filterChanged or 
                    (filterChanged and 
                     repo.isUnderVersionControl(item) and
                     repo.isResourceChanged(item))):
                    changed.insert(path)
                return True
            wt._recurseAndApply(self, collect)
        return changed

    def getLastAddedAndRemoved(self):
        wt = getToolByName(self, 'portal_workspaces')
        stage_name = wt.getStageOf(self)
        base_path = '/'.join(self.getPhysicalPath())

        removes = OITreeSet()
        adds = OITreeSet()

        # Get the list of current paths below this stage from the
        # catalog, and compare it with the last current list of paths
        # for this stage, to find out what has been deleted and added.
        applyFilter = bool(self.getDeployStates())
        curr = self.getDeploymentCandidates(applyFilter=applyFilter, 
                                            filterChanged=False,
                                            adds=adds)

        # Fetch the list of paths currently deployed by looking at the
        # last version path info structure.
        vpi = wt.getLastVersionPathInfo(self, stage_name)
        tree = vpi.paths(base_path)

        # The list of current content needs further filtering: if
        # something is marked as deployment candidate but it's parent
        # is not in the current list then it needs to be removed
        # (because the parent will be removed too). One such case is
        # caused by the parent folder being expired.
        seen = OITreeSet()
        checked = OITreeSet([base_path])
        skip = OITreeSet()
        for path in curr:
            parent = posixpath.dirname(path)
            if checked.has_key(parent):
                continue
            if seen.insert(parent): # Means it was not in seen.
                if (curr.has_key(parent) and not 
                    (skip.has_key(parent) or removes.has_key(parent))):
                    checked.insert(parent)
                    continue
            if tree.has_key(path):
                removes.insert(path)
            else:
                skip.insert(path)

        # If something was changed and didn't exist before, then it
        # means it was added.
        adds = union(adds, o_difference(curr, tree))

        # If it was added and already existed, means it was replaced
        # (thus, removed + added)
        removes = union(removes, o_intersection(adds, tree))

        # If something existed but doesn't exist anymore,
        # it was removed. Or?
        removes = union(removes, o_difference(tree, curr))

        # The list of adds needs further filtering: if something is
        # marked as added but it's parent is neither in adds or tree,
        # then it cannot be added (because the parent does not
        # exist). One such case is caused by the parent folder being
        # expired.
        seen = OITreeSet()
        checked = OITreeSet([base_path])
        for path in adds:
            parent = posixpath.dirname(path)
            if checked.has_key(parent):
                continue
            if seen.insert(parent): # Means it was not in seen.
                if ((adds.has_key(parent) or tree.has_key(parent)
                     and not skip.has_key(parent))):
                    checked.insert(parent)
                    continue
            skip.insert(path)

        adds = o_difference(adds, skip)

        try:
            adds.remove(base_path)
        except KeyError:
            pass
        try:
            removes.remove(base_path)
        except KeyError:
            pass
        return removes, adds

    def getLastMovements(self):
        index, annotLast = self.getLastReversibleAnnotation()
        index, annotPrevious = self.getLastReversibleAnnotation(before=index)
        wt = getToolByName(self, 'portal_workspaces')
        stage = wt.getStageOf(self)
        historyIdMapPrevious = wt.getPath2HistoryIdMap(
            stage, annotPrevious['args']['label'])
        historyIdMapLast = wt.getPath2HistoryIdMap(
            stage, annotLast['args']['label'])
        differ = SequenceMatcher(None, historyIdMapPrevious, historyIdMapLast)
        changes = differ.get_opcodes()
        removes, adds = interpretOpcodes(changes,
                                         historyIdMapPrevious,
                                         historyIdMapLast)
        removes = ["/".join(path) for path, histId in removes]
        adds = ["/".join(path) for path, histId in adds]
        return removes, adds

    def getLastMovementsAndModifications(self):
        index, annotLast = self.getLastReversibleAnnotation()
        index, annotPrevious = self.getLastReversibleAnnotation(before=index)
        return self.getMovementsAndModifications(
            annotPrevious['args']['label'],
            annotLast['args']['label'])

    def getMovementsAndModifications(self, labelPrevious, labelLast):
        # An object is considered:
        #
        # - Removed if a (path, history_id) tuple
        #   is present in the previous label but not on this one
        #
        # - Added if the inverse happens
        #
        # - Changed if the (path, history_id) remains
        #   between labels, but the version_id changes
        #
        wt = getToolByName(self, 'portal_workspaces')
        stage = wt.getStageOf(self)

        # Normalize labels.
        labelPrevious = self._globalLabel(labelPrevious)
        labelLast = self._globalLabel(labelLast)

        # Compute version info map for previous label
        vInfoMapPrevious = wt.getPath2VersionInfoMap(
            stage, labelPrevious)

        # Create a dictionary to ease our job computing modifications.
        historyIdMapPrevious = dict(
            [((path, history_id), version_id)
             for path, history_id, version_id in vInfoMapPrevious])
        himPreviousSet = Set(historyIdMapPrevious.keys())

        vInfoMapLast = wt.getPath2VersionInfoMap(
            stage, labelLast)
        historyIdMapLast = dict(
            [((path, history_id), version_id)
             for path, history_id, version_id in vInfoMapLast])
        himLastSet = Set(historyIdMapLast.keys())

        # Calculate the movements
        removes = himPreviousSet - himLastSet
        adds = himLastSet - himPreviousSet
        # Now calculate the modifications:
        # & -> intersection for sets
        himModSet = himLastSet & himPreviousSet

        # We've reduced the space to search for modifications by doing
        # an intersection with the . That should give us a small speed win. It
        # also makes sure there's no dupes in the 'modifications' set
        # below.
        versionIdMapPrevious = []
        versionIdMapLast = []
        for key in himModSet:
            path, history_id = key
            versionIdMapPrevious.append((path, historyIdMapPrevious[key]))
            versionIdMapLast.append((path, historyIdMapLast[key]))
        vimPreviousSet = Set(versionIdMapPrevious)
        vimLastSet = Set(versionIdMapLast)

        # Now, we calculate the real modifications by checking if the
        # version_id changed between the last and previous sets.
        modifications = vimLastSet - vimPreviousSet
        removes, adds, modifications = [_getPathAndSort(seq) for seq
                                        in removes, adds, modifications]
        return dict(removes=removes, adds=adds, modifications=modifications)

    security.declareProtected(StageObjects, 'publishLastChanged')
    def publishLastChanged(self, label=None, message=None, queued=False):
        transaction.savepoint(optimistic=True)
        if queued:
            # delegate to JobServer if present
            self.queueCall('publishLastChanged', label=label,
                           message=message)
            return

        stage_path = "/".join(self.getPhysicalPath())
        transaction.get().note(
            "Incremental deployment of stage '%s'" % stage_path)

        # now let's locate the changed objects...
        changedPaths = self.getLastChangedPaths(applyFilter=True)
        removes, adds = self.getLastAddedAndRemoved()
        return self.publishSelectedChanges(label=label, message=message, 
                                           queued=queued, adds=adds,
                                           removes=removes, 
                                           changedPaths=changedPaths)

    security.declareProtected(StageObjects, 'publishSelectedChanges')
    def publishSelectedChanges(self, label=None, message=None, queued=False,
                               adds=None, removes=None, changedPaths=None):
        publishStart(self, {}, update=True) # Notify the progress
                                            # manager of start
        if queued:
            # delegate to JobServer if present
            self.queueCall('publishSelectedChanges', label=label,
                           message=message, adds=adds, removes=removes, 
                           changedPaths=changedPaths)
            return


        if adds is None:
            adds = OITreeSet()
        if not isinstance(adds, OITreeSet):
            adds = OITreeSet(adds)
        if removes is None:
            removes = ()
        if changedPaths is None:
            changedPaths = ()

        if not (adds or removes or changedPaths):
            raise DeploymentError('Nothing was selected for deployment')

        # throw in the added objects along with changed and make it
        # one big list of things to be marked as changed.
        for path in changedPaths:
            adds.insert(path)
        changedPaths = list(adds)

        # compute some preliminary information about this deployment
        if not label:
            label = self.getNextLabel()

        globalLabel = self._globalLabel(label)
        tag_only = self.doInPlace()

        args = {'label':globalLabel,
                'message':message,
                'to_stage':None,
                'to_stage_path':None}

        info = {'operation': tag_only and 'tag' or 'partialPublish+tag',
                'args':args,
                'count':len(changedPaths), # only an aproximation
                'label': label}

        publishStart(self, info, update=True) # Notify the progress manager of
                                              # start, updating info.

        # checkin all modified objects
        wt = getToolByName(self, 'portal_workspaces')
        wt.checkInObjects([self.unrestrictedTraverse(path)
                           for path in changedPaths],
                          message=message,
                          stage=self.getSourceStageName())

        to_stages = None
        if not tag_only:
            si = wt.getStagesInfo()
            to_stage = si.get(self.getStagePath())

            __traceback_info__ = (self.getId(), 'destination stage: ' +
                                  str(to_stage), changedPaths, si)

            if to_stage is None:
                raise ValueError("Destination stage is not set. "
                                 "Cannot proceed.")
            to_stages = [to_stage[0]]

            # It's a publishing operation, update args info.
            args['to_stage'] = to_stage
            args['to_stage_path'] = self.getStagePath()

        # now, let's snapshot the current state so that we can
        # revert to it later, and calculate the moved objects
        # (removes, copies, renames)
        self.saveStage(label, message, info,
                       changed_paths=changedPaths,
                       removed_paths=removes)
        if tag_only:
            # Great, that saves us space and time!
            return

        # now let's get the movements
        removes, adds = self.getLastMovements()

        # find the duplicates between the changed objects and the
        # moved/copied ones
        pubPath = "/".join(self.getPhysicalPath())
        adds = [pubPath + '/' + path for path in adds]
        addSet = Set(adds)
        addSet.union_update(changedPaths)

        # let's update the newly added and changed objects
        paths = [path.split('/') for path in addSet]
        paths.sort()
        objs = [(self.unrestrictedTraverse(path), path) for path in paths]

        # now, filter the added objects to contain only objects in the
        # required states, if any state has been specified.
        states = self.getDeployStates()
        state_remove = []
        if states:
            states = tuple(states) + (None,)
            wf = getToolByName(self, 'portal_workflow')
            for obj, path in objs:
                state = wf.getInfoFor(obj, 'review_state', default=None)
                if state not in states:
                    # mark for removal if not in required state.
                    state_remove.append(path)

        # get the final list of objects.
        objs = [obj for obj, path in objs if path not in state_remove]

        # and the final list of removes
        l_pubPath = len(pubPath.split('/'))
        for path in state_remove:
            rel_path = path[l_pubPath:]
            if rel_path not in removes:
                removes.append('/'.join(rel_path))

        assert to_stages is not None

        # let's remove what's in the way...
        self._removeFromDestination(removes)

        # and finally publish the modified objects.
        wt.publishObjects(objs, to_stages)

    def _removeFromDestination(self, paths):
        # remove a bunch of relative paths from the destination stage
        root = getToolByName(self, 'portal_url').getPortalObject()
        dest_stage = root.unrestrictedTraverse(self.getStagePath())
        marker = object()
        for path in paths:
            obj = dest_stage.unrestrictedTraverse(path, marker)
            if obj is not marker:
                obj.aq_parent.manage_delObjects([obj.getId()])

    security.declareProtected(StageObjects, 'publishStage')
    def publishStage(self, label=None, message=None,
                     wipe=True, hardWipe=False,
                     queued=False):
        """Publish this stage to the targeted deployment stage.
        """
        publishStart(self, {}, update=True)
        transaction.savepoint(optimistic=True)
        if queued:
            # delegate to JobServer if present
            self.queueCall('publishStage', label=label, message=message,
                           wipe=wipe, hardWipe=hardWipe)
            return

        stage_path = "/".join(self.getPhysicalPath())
        transaction.get().note("full deployment of stage '%s'" % stage_path)
        wt = getToolByName(self, 'portal_workspaces')
        if not label:
            label = self.getNextLabel()

        globalLabel = self._globalLabel(label)

        tag_only = self.doInPlace()
        args = {'label':globalLabel,
                'message':message,
                'to_stage':None,
                'to_stage_path':None}
        info = {'operation': tag_only and 'tag' or 'publish+tag',
                'args':args,
                'count':self.count(),
                'label': label}

        if tag_only:
            # Great! Saves us space and time.
            publishStart(self, info, update=True)
            self.saveStage(label, message, info, force_tag=True)
            return
        
        si = wt.getStagesInfo()
        to_stage = si.get(self.getStagePath())

        __traceback_info__ = (self.getId(), 'destination stage' +
                              str(to_stage), si)

        if to_stage is None:
            raise ValueError, "Destination stage is not set. Cannot proceed."

        # It's a publishing operation, update args info.
        args['to_stage'] = to_stage
        args['to_stage_path'] = self.getStagePath()

        # Save information about deployment
        publishStart(self, info, update=True)
        self.annotate(info)

        to_stages = [to_stage[0]]
        source = self
        from_stage = wt.getStageOf(source)
        assert from_stage is not None
        wt.recursivePublish(source, from_stage, to_stages,
                            label=globalLabel, message=message,
                            wipe=wipe, hardWipe=hardWipe)

    security.declarePublic('isQueueingEnabled')
    def isQueueingEnabled(self):
        jqueue = self.unrestrictedTraverse('/job_queue', None)
        return (IJobQueue is not None and
                IJobQueue.providedBy(jqueue))

    security.declarePrivate('queueCall')
    def queueCall(self, callname, *args, **kw):
        assert self.isQueueingEnabled(), "JobServer not present. Can't queue call"
        jqueue = self.unrestrictedTraverse('/job_queue')

        methodpath = '/'.join(self.getPhysicalPath()) + '/' + callname
        user = getSecurityManager().getUser()
        username = user.getUserName()
        userpath = "/".join(user.aq_parent.getPhysicalPath())
        dispatch = jqueue.manage_addProduct['JobServer']
        jr = dispatch.manage_addJobRequest(methodpath=methodpath,
                                           username=username,
                                           userpath=userpath,
                                           args=args,
                                           kw=kw)
        return jr

    security.declareProtected(StageObjects, 'saveStage')
    def saveStage(self, label, message=None, info=None,
                  force_tag=False, changed_paths=None,
                  removed_paths=None):
        """Save and label this stage
        """
        publishStart(self, info, update=True) # Notify the progress manager of
                                              # start, only if not done yet.
        transaction.savepoint(optimistic=True)
        wt = getToolByName(self, 'portal_workspaces')

        __traceback_info__ = (self.getId(),)

        globalLabel = self._globalLabel(label)
        source = self
        from_stage = wt.getStageOf(source)
        assert from_stage is not None
        if changed_paths is None:
            wt.recursiveTag(source, from_stage, label=globalLabel,
                            message=message, force_tag=force_tag)
        else:
            wt.incrementalTag(source, from_stage, label=globalLabel,
                              message=message,
                              changed_paths=changed_paths,
                              removed_paths=removed_paths)

        if info is None:
            args = {'label': globalLabel,
                    'message': message}
            info = {'operation': 'tag',
                    'args': args,
                    'count': self.count(),
                    'label': label}
        self.annotate(info)

    security.declarePublic('isReversibleAnnotation')
    def isReversibleAnnotation(self, annotation):
        """check if the annotation represents a reversible tag"""
        # 'tag' must be present in the annotation operation
        # and the global label must be valid
        wt = getToolByName(self, 'portal_workspaces')

        return ('tag' in annotation['operation'] and
                wt.isValidLabel(self._getSourceStageName(),
                                annotation['args']['label']))

    security.declareProtected(StageObjects, 'updateStage')
    def updateStage(self, label):
        """Update stage to a given label.
        """
        # label here is a globalLabel
        if not label:
            raise ValueError, 'Cannot revert to empty label'
        if label not in [annotation['args']['label']
                         for annotation in self.getAnnotations()
                         if self.isReversibleAnnotation(annotation)
                         ]:
            raise ValueError, 'Cannot revert to label "%s"' % label
        wt = getToolByName(self, 'portal_workspaces')
        source = self
        wt.recursiveRevertToLabel(source, label)
        args = {'label':label}
        annotation = {'operation': 'update',
                      'count':self.count(),
                      'args':args}
        self.annotate(annotation)

    security.declarePublic('ess_isUnderWorkspace')
    ess_isUnderWorkspace = True

    security.declareProtected(StageObjects, 'getCurrentChangeSet')
    def getCurrentChangeSet(self):
        """ Informational method that returns a tree of
        modified/added/deleted objects since the last deployment.  
        
        This method should not be taken as authoritative, though it
        gets pretty close.
        """

        # let's locate the changed objects...
        changedPaths = self.getLastChangedPaths(applyFilter=True)
        removes, adds = self.getLastAddedAndRemoved()

        status_map = dict([(tuple(path.split('/')),
                            dict(path=path,
                                 id=path.split('/')[-1],
                                 url=None,
                                 status='removed'))
                           for path in removes])

        ct = getToolByName(self, 'portal_catalog')
        ut = getToolByName(self, 'portal_url')
        portal = ut.getPortalObject()

        def get_info(path):
            try:
                item = ct.getMetadataForUID(path)
                info = {'title': item['title_or_id'],
                        'portal_type': item['portal_type']}
            except KeyError:
                item = portal.unrestrictedTraverse(path, None)
                if item is None:
                    return {}
                pt = getattr(aq_base(item), 'portal_type', None)
                if pt is None:
                    pt = item.meta_type
                info = {'title': item.title_or_id(),
                        'portal_type': pt}
            return info

        path2url = self.REQUEST.physicalPathToURL
        for status, paths in (('added', adds),
                              ('changed', changedPaths)):
            for path in paths:
                split_path = tuple(path.split('/'))
                info = get_info(path)
                info.update({'path': path,
                             'id': split_path[-1],
                             'url': path2url(path),
                             'status': status})
                status_map[split_path] = info
        status_map = status_map.items()
        status_map.sort()
        return [info for path, info in status_map]

    security.declarePrivate('annotate')
    def annotate(self, info):
        mt = getToolByName(self, 'portal_membership')
        more_info = {'user':mt.getAuthenticatedMember().getId(),
                     'date':DateTime()}
        info.update(more_info)
        self._si_length.change(1)
        next = self._si_length()
        self._stage_info[next] = info

    security.declarePublic('getLastAnnotation')
    def getLastAnnotation(self):
        """ Get the last annotation
        """
        return self._stage_info[self._si_length()]

    security.declarePublic('getLastReversibleAnnotation')
    def getLastReversibleAnnotation(self, before=None):
        """ Get the last annotation that is also a tag operation
        """
        if before is None:
            start = self._si_length()
        else:
            start = before - 1
        for i in xrange(start, 0, -1):
            annot = self._stage_info[i]
            if self.isReversibleAnnotation(annot):
                return i, annot
        raise ValueError, "No remaining reversible operations"

    security.declarePublic('mayPublishIncrementally')
    def mayPublishIncrementally(self):
        try:
            self.getLastReversibleAnnotation()
        except ValueError:
            return False
        else:
            return True

    security.declarePublic('getAnnotations')
    def getAnnotations(self, count=None):
        """ Get a sequence of annotations.

        If the count argument is provided, the 'count'
        last annotations are returned.
        """
        start = 1
        last = self._si_length()
        if count is not None:
            start = last - count
        return [self._stage_info[i] for i in range(start, last+1)]

    security.declarePrivate('count')
    def count(self):
        res = Length()
        def add(obj):
            res.change(1)
        recurseVisit(self, add)
        return res()

    security.declareProtected(StageObjects, 'workspace_publish_stage')
    def workspace_publish_stage(self, IncrementalDeployment=False, 
                                label=None, comment='', queued=False):
        """ Publish a staging area
        """
        publishStart(self, {}) # Notify progress manager of start.
        request = self.REQUEST
        response = request.RESPONSE

        url = '%s/workspace_deployment' % self.absolute_url()

        catalog = self.portal_catalog
        def processQueueCatalog():
            process = getattr(catalog, 'process', None)
            if process:
                items = catalog.getIndexInfo()
                process(len(items))

        processQueueCatalog()

        added = request.get('added')
        removed = request.get('removed')
        changed = request.get('changed')

        msg = None
        if label is not None:
            for info in self.getAnnotations():
                if info.get('label', '').strip() == label.strip():
                    msg = ('The label, %s has already been used.  '
                           'Please select another.' % str(label))
                    break
        else:
            label = self.getNextLabel()

        if msg is None:
            if queued:
                msg = "Queued "
            else:
                msg = "Completed "
            try:
                if IncrementalDeployment:
                    if added or removed or changed:
                        self.publishSelectedChanges(
                            label=label, message=comment, 
                            adds=added,
                            removes=removed,
                            changedPaths=changed,
                            queued=queued)
                    else:
                        self.publishLastChanged(label=label, message=comment, 
                                                queued=queued)
                    msg += 'Incremental'
                else:
                    self.publishStage(label=label, message=comment, 
                                      queued=queued)
                    msg += 'Full'
            except (DeploymentError, ClientDisconnected, ConnectionError), e:
                try:
                    log = aq_acquire(self, '__error_log__', containment=1)
                    log.raising(sys.exc_info())
                except AttributeError:
                    pass

                errormsg = ''
                if isinstance(e, URLRewritingError):
                    args = tuple(e.args[:4])
                    errormsg = (
                        'Could not find enough information'
                        'about object with UID "%s" to rewrite the '
                        'URL %s while deploying %s (context: %s).' % args)
                elif isinstance(e, (DeploymentInProgress, OSError, 
                                    DependencyError)):
                    errormsg = str(e)
                
                elif isinstance(e, ClientDisconnected):
                    errormsg = 'Entransit target is not running'
                
                elif isinstance(e, ConnectionError):
                    args = tuple(e.args[:2])
                    errormsg = 'Connection error, status %s, message %s' % args
                elif isinstance(e, (RemoteDeploymentError, DeploymentError)):
                    errormsg = str(e)
                # Since we are catching the exception here instead of
                # letting it bubble up, we have to abort the
                # transaction ourselves, otherwise changes might end
                # up being commited!
                transaction.abort()
                return self.workspace_deployment_error(message=errormsg)
            msg += " Deployment of " + self.title_or_id()

        processQueueCatalog()
        
        query = "?" + urlencode(dict(portal_status_message=msg))
        response.redirect(url + query)
        return url + query

def recurseVisit(obj, func):
    for item in obj.objectValues():
        func(item)
        if item.aq_explicit.isPrincipiaFolderish:
            recurseVisit(item, func)
    return

def interpretOpcodes(opcodes, ilist, jlist):
    removes = []
    adds = []
    for tag, i1, i2, j1, j2 in opcodes:
        if tag == 'equal':
            # we're not interested in the areas that match
            continue
        if tag == 'delete' or tag == 'replace':
            removes.extend(ilist[i1:i2])
        if tag == 'insert' or tag == 'replace':
            adds.extend(jlist[j1:j2])
    return removes, adds

def _getPathAndSort(sequence):
    sequence = [path for path, dummy in sequence]
    sequence.sort()
    return sequence

registerType(StagingArea, PROJECTNAME)

# XXX This is not used anymore. But we will leave it around so pickles
# don't break.
class Sandbox(OrderedBaseFolder):
    """ Staging Sandbox """

    meta_type = portal_type = archetype_name = 'Sandbox'
    global_allow = 0

# Don't register it anymore.
# registerType(Sandbox, PROJECTNAME)
