# EnSimpleStaging  |  Copyright(C), 2008, Enfold Systems, Inc.
# see LICENSE.txt for details

import simplejson
from copy import deepcopy
from logging import getLogger
from zope.interface import implements

import transaction
from transaction._transaction import Status
from transaction.interfaces import ISavepointDataManager, ISynchronizer

from DateTime import DateTime
from Acquisition import aq_inner, aq_parent, aq_base

from Products.EnSimpleStaging.event import subscribe
from Products.EnSimpleStaging.interfaces import IExceptionLoggedEvent
from Products.EnSimpleStaging.interfaces import IObjectStagedEvent
from Products.EnSimpleStaging.interfaces import IObjectTaggedEvent
from Products.EnSimpleStaging.interfaces import IObjectCheckInEvent
from Products.EnSimpleStaging.interfaces import IPublishStageStartEvent
from Products.EnSimpleStaging.interfaces import IPublishStageFailureEvent
from Products.EnSimpleStaging.interfaces import IPublishStageFinishEvent
from Products.EnSimpleStaging.event import publishFinish, publishFailure

logger = getLogger('EnSimpleStaging.progress')

STARTED = 'Started'
WORKING = 'Working'
ERROR = 'Error'
FINISHED = 'Finished'
UNKNOWN = 'Unknown'

# Why are we using a DataManager *and* a Synchronizer for such a
# 'simple' functionality?
#
# It might feel like overkill, or that it's overly complex to you. But
# there are very concrete reasons for the code to be structured this
# way! So hold your breath, delete that email you were about to fire
# and read on.
#
# We are maintaining global state that needs to be consulted by other
# threads here. Specially, this code is going to be called repeatedly
# (polling) by some browser-side Javascript so the answer needs to be
# very accurate because the Javascript will most certainly fire other
# actions depending on what we say the current progress status is!
#
# More importantly, our state needs to be accurate in respect to:
#
#  a) Savepoints: if a savepoint is used and affects the count of
#  objects that have been deployed, we want the count of deployed
#  objects to stay correct. For this reason we register a DataManager
#  that supports savepoints.
#
#  b) Transaction commit/abort: It is only safe to say that the
#  deployment has failed/succeded (and thus, that other
#  threads/requests will get correct data) after the
#  transaction.status is COMMITTED. The best way to do that (since we
#  don't commit any data ourselves) is to use the
#  ISynchronizer.afterCompletion() functionality, which is called very
#  late, further after the DataManager commit()/abort() have been
#  called and also exists in older Zope versions. So late actually
#  that the transaction has been released by the TransactionManager!
#
# Hopefully that explains any confusion that you might have.

class PDMSavepoint(object):
    """ Support for transaction savepoints.
    """

    def __init__(self, progress, stage_id):
        self.progress = progress
        self.stage_id = stage_id
        self.state = progress.getState(self.stage_id)

    def rollback(self):
        self.progress.setState(self.stage_id, self.state)

class ProgressDataManager(object):
    """ Data Manager for keeping track of deployment progress
    """
    implements(ISavepointDataManager, ISynchronizer)

    def __init__(self, progress, stage_id):
        logger.debug('initializing %r (stage %r).', 
                     self, stage_id)
        self.registered = False
        self.vote = False
        self.progress = progress
        self.stage_id = stage_id

    # IDataManager API
    def register(self):
        if not self.registered:
            try:
                _transaction = transaction.get()
                _transaction.join(self)
                transaction.manager.registerSynch(self)
                self.registered = True
                logger.debug('registered %r (stage %r).', self, self.stage_id)
            except:
                logger.exception('exception during register (registered=%s)', 
                                 self.registered)

    def savepoint(self):
        return PDMSavepoint(self.progress, self.stage_id)

    def tpc_begin(self, _transaction):
        pass

    def commit(self, _transaction):
        pass

    def tpc_vote(self, _transaction):
        pass

    def tpc_finish(self, _transaction):
        self.registered = False

    def tpc_abort(self, _transaction):
        pass

    abort = tpc_abort

    def sortKey(self):
        return id(self)

    # ISynchronizer API
    def afterCompletion(self, _transaction):
        transaction.manager.unregisterSynch(self)
        if not _transaction.status == Status.COMMITTED:
            publishFailure(self.stage_id)
        else:
            publishFinish(self.stage_id)

    def beforeCompletion(self, _transaction):
        pass

    def newTransaction(self, _transaction):
        pass

class DeploymentProgress(object):
    
    def __init__(self):
        self.stages = {}

    def getState(self, stage_id):
        if not stage_id in self.stages:
            return {'status': UNKNOWN}
        return deepcopy(self.stages[stage_id])

    def setState(self, stage_id, state):
        self.stages[stage_id] = state

    def start(self, stage, info, update=False):
        stage_id = stage._getSourceStageName()

        if stage_id in self.stages and not update:
            if self.stages[stage_id]['status'] not in (FINISHED, ERROR):
                # Let's just log a warning here. This is not such a
                # critical condition. Moreover, it would require every
                # test to do a full commit() just to satisfy this
                # condition.
                #
                # raise ValueError('Deployment of %r is already in progress' 
                #                  % stage_id)
                logger.warning('Deployment of %r is already in progress',
                               stage_id)

        if update and stage_id in self.stages:
            # Just update the information about the current
            # deployment with the most recent info.
            if info:
                self.stages[stage_id]['info'] = info
        else:
            self.stages[stage_id] = {'status': STARTED,
                                     'title': stage.Title(),
                                     'process': {},
                                     'info': info}
    
        ProgressDataManager(self, stage_id).register()

    def tickle(self, stage_id, current, process):
        self.stages[stage_id]['status'] = WORKING
        if not self.stages[stage_id]['process'].has_key(process):
            self.stages[stage_id]['process'][process] = {
                'count': 0,
                'current': ''}
        self.stages[stage_id]['process'][process]['count'] += 1
        self.stages[stage_id]['process'][process]['current'] = current
    
    def error(self, stage_id):
        self.stages[stage_id]['status'] = ERROR

    def finish(self, stage_id):
        self.stages[stage_id]['status'] = FINISHED

PROGRESS = DeploymentProgress()

class DateTimeEncoder(simplejson.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, DateTime):
            return float(obj)
        return simplejson.JSONEncoder.default(self, obj)

def getStatusFor(stage_id, format=None):
    state = PROGRESS.getState(stage_id)
    if format is None:
        return state
    elif format == 'JSON':
        from simplejson import dumps
        return dumps(state, cls=DateTimeEncoder)
    raise ValueError('Unknown format requested: %r' % format)

def deployObjectSubscriber(event):
    assert event.stage is not None, 'Missing stage in IDeployObjectEvent'
    stage_id = event.stage
    current = event.source.absolute_url()
    PROGRESS.tickle(stage_id, current, 'DEPLOY')
subscribe(IObjectStagedEvent, deployObjectSubscriber)

def taggedObjectSubscriber(event):
    assert event.stage is not None, 'Missing stage in IDeployObjectEvent'
    stage_id = event.stage
    current = event.source.absolute_url()
    PROGRESS.tickle(stage_id, current, 'TAG')
subscribe(IObjectTaggedEvent, taggedObjectSubscriber)

def checkinObjectSubscriber(event):
    assert event.stage is not None, 'Missing stage in IDeployObjectEvent'
    stage_id = event.stage
    current = event.source.absolute_url()
    PROGRESS.tickle(stage_id, current, 'CHECKIN')
subscribe(IObjectCheckInEvent, checkinObjectSubscriber)

def exceptionLoggedSubscriber(event):
    stage_id = None
    entry_id = event.entry_id
    error_log = event.source
    request = getattr(error_log, 'REQUEST', None)
    if request is not None:
        # Try to find the current stage from the request's published
        # object.
        parents = request.get('PARENTS', ())
        for parent in parents:
            if getattr(aq_base(parent), 'meta_type', None) == 'Staging Area':
                stage_id = parent.getSourceStageName()
                break
    if stage_id is None:
        # Try to guess the current stage being published by
        # looking at the stages which are not in final states.
        #
        # Usually this will not happen because we will be able to
        # figure the stage from request['PUBLISHED'] above.
        candidates = []
        for key, item in PROGRESS.stages.items():
            if item['status'] not in (FINISHED, ERROR):
                candidates.append(key)
        if len(candidates) == 1:
            stage_id = candidates[0]
        else:
            # There was more than one stage being deployment at the
            # same time, and we could not figure out which one
            # triggered this error.
            return
    current = error_log.getLogEntryById(entry_id)
    for key in ('tb_html', 'req_html'):
        try:
            del current[key]
        except KeyError:
            pass
    PROGRESS.tickle(stage_id, current, 'ERROR')
subscribe(IExceptionLoggedEvent, exceptionLoggedSubscriber)

def publishStageStart(event):
    PROGRESS.start(event.stage, event.info, update=event.update)
subscribe(IPublishStageStartEvent, publishStageStart)

def publishStageFinish(event):
    PROGRESS.finish(event.stage)
subscribe(IPublishStageFinishEvent, publishStageFinish)

def publishStageFailure(event):
    PROGRESS.error(event.stage)
subscribe(IPublishStageFailureEvent, publishStageFailure)
