##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################

__version__='$Revision: 1.11 $'[11:-2]

import time
import tempfile
import logging

from sets import Set
from itertools import chain
from cPickle import Pickler, Unpickler

from Acquisition import Implicit, aq_parent, aq_inner, aq_base
from Globals import InitializeClass, Persistent
from AccessControl import ClassSecurityInfo
from BTrees.OOBTree import OOBTree
from BTrees.OIBTree import OIBTree
from BTrees.Length import Length
from OFS.SimpleItem import SimpleItem

from ZODB.utils import p64
from ZODB.serialize import ObjectWriter, referencesf

from Utility import VersionControlError
from nonversioned import listNonVersionedObjects, removeNonVersionedData

logger = logging.getLogger('ZopeVersionControl')
export_end_marker = '\377'*16

class FakeList(Set):
    def append(self, item):
        self.add(item)

def exportObject(obj, oid, jar, load, version, ignore_dict):
    f = tempfile.TemporaryFile()
    f.write('ZEXP')

    oids = FakeList([oid])
    done_oids = Set()

    # We use a special ObjectWriter that modifies the object state to
    # be serialized right before it is actually serialized. By doing
    # so, we only serialize the state that is actually important and
    # removeNonVersionedData() has less work to do because the object
    # will be (most of the time) already clean.
    writer = CloningObjectWriter(ignore_dict)
    writer._jar = jar

    def serialize(obj):
        # 'writer' keeps a stack of the objects it sees when and then
        # when we iterate over it those objects are returned.
        for obj in chain([obj], writer):
            oid = obj._p_oid
            if oid in oids:
                oids.remove(oid)
            if oid in done_oids:
                continue
            done_oids.add(oid)
            try:
                state = obj._p_changed
            except AttributeError:
                state = 0
            
            new_p = writer.serialize(obj)

            if state is None:
                obj._p_deactivate()

            referencesf(new_p, oids)
            f.writelines([oid, p64(len(new_p)), new_p])

    # Serialize anything that is directly reachable from the object.
    serialize(obj)

    while oids:
        oid = oids.pop()
        if oid in done_oids:
            continue
        
        obj = jar.get(oid) # XXX Should we handle broken references
                           # here? A KeyError most likely means a new
                           # object that wasn't registered with the
                           # connection yet.
        serialize(obj)
        
    f.write(export_end_marker)
    return f

def cloneByPickle(obj, ignore_list=(), to_jar=None):
    """Makes a copy of a ZODB object, loading ghosts as needed.

    Ignores specified objects along the way, replacing them with None
    in the copy.
    """
    ignore_dict = {}
    for o in ignore_list:
        ignore_dict[id(o)] = o

    # Originally this code used StringIO() to store the pickle, and
    # used plain cPickle, however since StringIO() is an in-memory
    # stream, we would end up with 3 copies of the data in memory,
    # albeit for a short period of time: the original object, the
    # pickled data and the newly-copied object. To avoid this we've
    # switched to using a temporary file, which the implementation
    # ~30% slower but using ~30% less memory.
    assert getattr(obj, '_p_oid', None) is not None, repr(obj)
    assert getattr(obj, '_p_jar', None) is not None, repr(obj)

    oid = obj._p_oid
    jar = obj._p_jar
    if to_jar is None:
        to_jar = jar
    load = jar._storage.load
    version = jar._version

    # Using the 'export to ZEXP' feature with the jar's importFile()
    # avoids a recursion error that we were getting when using cPickle
    # because it serializes each object's state separatedly instead of
    # serializing the whole object and all its dependencies in one big
    # blob.
    f = exportObject(obj, oid, jar, load, version, ignore_dict)
    f.seek(0)
    return to_jar.importFile(f)

class CloningObjectWriter(ObjectWriter):

    def __init__(self, ignore_dict):
        ObjectWriter.__init__(self)
        self.ignore_dict = ignore_dict

    def _dump(self, meta, state):
        if state is not None and isinstance(state, dict):
            # Do the removal of non-versioned data here to ease the
            # job of removeNonVersionedData() below. This is slightly
            # smarter than the previous approach of cloning the whole
            # object using cPickle and then removing the non-versioned
            # attributes because we are now pickling only the data
            # that's actually important.
            if state.has_key('__vc_ignore__'):
                for k in state['__vc_ignore__']:
                    if state.has_key(k):
                        del state[k]
            removed = {}
            if self.ignore_dict:
                for k, v in state.iteritems():
                    if id(v) in self.ignore_dict:
                        removed[k] = None
                for k in removed.iterkeys():
                    del state[k]
            if state.has_key('_objects'):
                # It's a folderish thingie, update it's _objects attribute.
                state['_objects'] = tuple([info for info in state['_objects']
                                           if not removed.has_key(info['id'])])
            if state.has_key('_tree') and state.has_key('_mt_index'):
                # Smells like a BTreeFolder-based thingie.
                tree = state['_tree']

                nt = state['_tree'] = OOBTree()
                nmt = state['_mt_index'] = OOBTree()
                nl = state['_count'] = Length()

                for name, object in tree.items():
                    if id(object) in self.ignore_dict:
                        continue

                    nt[name] = object
                    nl.change(1)
                    # Update the meta type index.
                    meta_type = getattr(object, 'meta_type', None)
                    if meta_type is not None:
                        ids = nmt.get(meta_type, None)
                        if ids is None:
                            ids = OIBTree()
                        nmt[meta_type] = ids
                        ids[name] = 1
                
        return ObjectWriter._dump(self, meta, state)


class Version(Implicit, Persistent):
    """A Version is a resource that contains a copy of a particular state
       (content and dead properties) of a version-controlled resource.  A
       version is created by checking in a checked-out resource. The state
       of a version of a version-controlled resource never changes."""

    def __init__(self, version_id, obj):
        self.id = version_id
        self.date_created = time.time()
        self._data = None

    # These attributes are set by the createVersion method of the version
    # history at the time the version is created. The branch is the name
    # of the branch on which the version was created. The prev attribute
    # is the version id of the predecessor to this version. The next attr
    # is a sequence of version ids of the successors to this version.
    branch = 'mainline'
    prev = None
    next = ()

    security = ClassSecurityInfo()

    security.declarePublic('getId')
    def getId(self):
        return self.id

    security.declarePrivate('saveState')
    def saveState(self, obj, container=None):
        """Save the state of object as the state for this version of
           a version-controlled resource."""
        if container is None:
            container = self
        self._data = self.stateCopy(obj, container)

    security.declarePrivate('copyState')
    def copyState(self, container=None):
        """Return an independent deep copy of the state of the version."""
        data = self.__dict__.get('_data')  # Avoid __of__ hooks
        if container is None:
            container = self
        return self.stateCopy(data, container)

    security.declarePrivate('stateCopy')
    def stateCopy(self, obj, container):
        """Get a deep copy of the state of an object.

        Breaks any database identity references.
        """
        to_jar = container._p_jar
        ignore = listNonVersionedObjects(obj)
        res = cloneByPickle(aq_base(obj), ignore, to_jar)
        if to_jar is not None:
            assert res._p_jar is to_jar, (res, res._p_jar, to_jar.connections)
        removeNonVersionedData(res)
        return res


InitializeClass(Version)
